import asyncio
from functools import cached_property
from typing import Generic, List, Type, TypeVar


class BaseElement:

    def __init__(self, size: int, allocated: bool):
        self.size = size
        self.allocated = allocated

    def is_allocated(self) -> bool:
        return self.allocated

    async def cleanup(self):
        self.allocated = False

    def allocate(self):
        """Pre consume the resource."""
        if self.allocated:
            return []
        self.allocated = True
        return [self]

    def is_unavailable(self) -> bool:
        return self.is_allocated()


T_co = TypeVar('T_co', bound=BaseElement)


class BaseResources(List[T_co]):

    def size(self):
        return sum(res.size for res in self if res.is_allocated())

    async def cleanup(self):
        for res in self:
            await res.cleanup()

    def __hash__(self) -> int:
        return hash(tuple(self))


class BaseAllocator(Generic[T_co]):

    pool: "BasePool"

    def __init__(self, pool: "BasePool"):
        self.pool = pool

    async def _get_size(self, size: int):
        """Get the size of the resources to allocate. Share the same signature as `_callback` except for the first argument."""
        return size

    async def _callback(self, _allocated: BaseResources, *args, **kwargs):
        """Do something after the resources are allocated. Share the same signature as `_get_size` except for the first argument."""
        pass

    async def _allocate(self, size: int):
        allocated = []
        for res in self.pool:
            allocated.extend(res.allocate())
            if sum(a.size for a in allocated) >= size:
                break
        return allocated

    async def __call__(self, *args, **kwargs) -> BaseResources[T_co]:
        allocated = BaseResources[T_co]()
        print_after = 5
        interval = 1

        while allocated.size() < (size := await
                                  self._get_size(*args, **kwargs)):
            if print_after == 0:
                print(f"Waiting for {size} {self.pool} resources...")
            if print_after >= 0:
                print_after -= 1
            await asyncio.sleep(interval)
            allocated.extend(await self._allocate(size))

        await self._callback(allocated, *args, **kwargs)
        return allocated


class BasePool:

    element_type: Type[BaseElement] = BaseElement
    pool: List[BaseElement]

    @cached_property
    def allocate(self):
        """Base allocator.

        Args:
            size: The size of the resources to allocate."""
        return BaseAllocator[BaseElement](self)

    def __iter__(self):
        return iter(self.pool)

    @property
    def available_size(self):
        return sum(res.size for res in self.pool if res.is_unavailable())



import asyncio
import subprocess
import threading
from logging import getLogger
from typing import Any, Dict, List, Set

from ..pools.base import BaseAllocator, BaseResources
from .runner import BaseRunner

logger = getLogger(__name__)


class Exp:

    def __init__(self, runner: BaseRunner, uuid: str):
        self.runner = runner
        self.uuid = uuid
        self.resources: Set[BaseResources] = set()

    async def get(self, alloc: BaseAllocator, *args, **kwargs):
        resource = await alloc(*args, **kwargs)
        self.resources.add(resource)
        return resource

    async def cleanup(self):
        for resource in self.resources:
            logger.debug(f"Cleaning up {resource}")
            await resource.cleanup()
        self.resources.clear()

    async def run(self, args: List[str], env: Dict[str, str], **kwargs) -> str:

        returns = None

        def run_in_thread(**popen_kwargs):
            proc = subprocess.Popen(**popen_kwargs)
            proc.wait()
            nonlocal returns
            returns = proc.returncode, proc.stdout.read().decode()
            return

        popen_kwargs = {
            "args": args,
            "env": env,
            "stdout": subprocess.PIPE,
        }

        thread = threading.Thread(target=run_in_thread,
                                  kwargs={
                                      **popen_kwargs,
                                      **kwargs
                                  })
        thread.start()

        while thread.is_alive():
            await asyncio.sleep(1)

        returncode, stdout = returns
        if returncode != 0:
            raise subprocess.CalledProcessError(returncode, args, stdout)

        return stdout

    async def report(self, metrics: Dict[str, Any], **kwargs):
        metrics.update(kwargs)
        await self.runner._report(self.uuid, metrics)



import re

import ml_scheduler

# at least 90% of the GPU memory should be free
cuda = ml_scheduler.pools.CUDAPool([0, 2], 90)
disk = ml_scheduler.pools.DiskPool('/one-fs')


@ml_scheduler.exp_func
async def mmlu(exp: ml_scheduler.Exp, model, checkpoint):
    """Copy the model from another filesystem and run inferences on one GPU."""

    source_dir = f"/another-fs/model/{model}/checkpoint-{checkpoint}"
    target_dir = f"/one-fs/model/{model}-{checkpoint}"

    # resources will be cleaned up after exiting the function
    disk_resource = await exp.get(
        disk.copy_folder,
        source_dir,
        target_dir,
        cleanup_target=True,
    )
    cuda_resource = await exp.get(cuda.allocate, 1)

    async def report(m, stdout):
        # parse the metrics from the output
        # See https://github.com/RUCAIBox/LLMBox for more details
        parts = stdout.rsplit("#####", 1)
        metrics = re.search(r"\d+\.\d+", parts[1]).group()
        await exp.report({m: metrics})

    # run inference
    args = [
        "python", "inference.py", "--model", target_dir, "--dataset", "mmlu"
    ]
    env = {
        "CUDA_DEVICE_ORDER": "PCI_BUS_ID",
        "CUDA_VISIBLE_DEVICES": ",".join(map(str, cuda_resource)),
    }
    stdout = await exp.run(args=args, env=env)
    await report('Accuracy', stdout)

    # run another inference with the same resources
    args = [
        "python", "inference.py", "--model", target_dir, "--dataset", "mmlu",
        "--num_shots", "5"
    ]
    stdout = await exp.run(args=args, env=env)
    await report('Accuracy (5-shots)', stdout)


mmlu.run_csv(
    "experiments.csv",
    ['Accuracy', 'Accuracy (5-shots)'],
)



"""**ml_scheduler**

A lightweight machine learning experiments scheduler in a few lines of simple Python
"""
__version__ = "1.0.0"
import coloredlogs

coloredlogs.install()

from . import pools
from .exp import Exp, exp_func
from .threads import to_thread

__all__ = "pools", "Exp", "exp_func", "to_thread"



from typing import Optional

from .base import BaseElement, BasePool


class CounterPool(BasePool):

    def __init__(
        self,
        size: int,
        available: Optional[int],
    ):
        self.size = size
        available = available if available is not None else size
        self.pool = [BaseElement(1, i >= available) for i in range(size)]



import asyncio
import os
import shutil
from logging import getLogger
from typing import Any, Dict, List, Optional
from uuid import uuid4

import pandas

from .base import BaseRunner

logger = getLogger(__name__)


class CSVRunner(BaseRunner):

    def run(
        self,
        csv_path: str,
        continue_cols: List[str],
        force_rerun: bool = False,
        read_csv_kwargs: Optional[dict] = None,
        uuid_column: str = ":uuid:",
        retval_column: Optional[str] = ":retval:",
    ):
        """Run experiments from a csv file

        Args:
            csv_path (`str`): The path to the csv file.
            continue_cols (`List[str]`): Run experiments where the columns are null.
            force_rerun (`bool`, optional): Force rerun all experiments. Ignore `continue_cols`. Defaults to False.
            read_csv_kwargs (`Optional[dict]`, optional): Additional kwargs passed to `pandas.read_csv`. Defaults to None.
            uuid_column (`str`, optional): The column name for the uuid. Defaults to `":uuid:"`.
            retval_column (`Optional[str]`, optional): The column name for the return value. None for not saving the return value. Defaults to `":retval:"`.
        """
        kwargs = {
            "csv_path": csv_path,
            "continue_cols": continue_cols,
            "force_rerun": force_rerun,
            "read_csv_kwargs": read_csv_kwargs or {},
            "uuid_column": uuid_column,
            "retval_column": retval_column,
        }
        return asyncio.run(self.arun(**kwargs))

    def submit_from_csv(
        self,
        force_rerun: bool = False,
    ):
        if os.path.exists(self.csv_path + ".lock"):
            os.remove(self.csv_path + ".lock")
        df: pandas.DataFrame = pandas.read_csv(self.csv_path,
                                               **self.read_csv_kwargs)

        # set uuid
        if self.uuid_column not in df.columns:
            df[self.uuid_column] = [str(uuid4()) for _ in range(len(df))]
        else:
            rows = df[self.uuid_column].isnull()
            df.loc[rows, self.uuid_column] = [
                str(uuid4()) for _ in range(len(df[rows]))
            ]

        df = df.set_index(self.uuid_column)
        df.to_csv(self.csv_path, index=True)

        # force rerun
        if not force_rerun:
            rows = False
            for col in self.continue_cols:
                rows |= df[col].isnull()
            added = int(rows.sum())
            logger.info(f"Adding {added} tasks ({len(df) - added} skipped).")
        else:
            rows = slice(None)
            logger.info(f"Adding {len(df)} tasks.")

        tasks = [
            self.create_task(uuid, **row) for uuid, row in df[rows].iterrows()
        ]

        return tasks

    async def __write_cell(self, row, col, value):

        lock_file = self.csv_path + ".lock"
        timeout = 10
        counts = 0
        while os.path.exists(lock_file):
            await asyncio.sleep(0.1)
            counts += 1
            if counts == timeout:
                logger.error(f"Timeout waiting for lock file {lock_file}")
                return

        shutil.copy(self.csv_path, lock_file)
        df = pandas.read_csv(self.csv_path, index_col=self.uuid_column)
        df.loc[row, col] = value
        df.to_csv(self.csv_path, index=True)
        os.remove(lock_file)

    async def _report(self, uuid: str, metrics: Dict[str, Any]):
        logger.info(f"Reporting {uuid} {metrics}")
        for metric, value in metrics.items():
            await self.__write_cell(uuid, metric, value)

    async def arun(
        self,
        csv_path: str,
        continue_cols: List[str],
        force_rerun: bool = False,
        read_csv_kwargs: Optional[dict] = None,
        uuid_column: str = ":uuid:",
        retval_column: Optional[str] = ":retval:",
    ):
        """Async run experiments from a csv file"""

        self.csv_path = csv_path
        self.continue_cols = continue_cols
        self.read_csv_kwargs = read_csv_kwargs
        self.uuid_column = uuid_column

        tasks = self.submit_from_csv(force_rerun)

        # block until all tasks are done
        for task in asyncio.as_completed(tasks):
            exp, results = await task
            logger.info(f"Finished {exp.uuid}")
            if retval_column is not None:
                await self.__write_cell(exp.uuid, retval_column, results)



from functools import cached_property
from typing import List

from nvitop import Device

from .base import BaseAllocator, BaseElement, BasePool


class CUDAElement(BaseElement):

    def __init__(self, device: Device, min_memory: float = 90):
        super().__init__(1, False)
        self.device = device
        self.min_memory = min_memory
        self.cuda_index = device.cuda_index

    def is_unavailable(self):
        return (100 - self.device.memory_percent()
                ) > self.min_memory or self.is_allocated()

    def __str__(self) -> str:
        return str(self.cuda_index)

    def __repr__(self) -> str:
        return "CUDA(device={}, allocated={}, memory_percent={})".format(
            self.device.cuda_index, self.allocated,
            self.device.memory_percent())


class CUDAPool(BasePool):

    def __init__(self, ids: List[int], min_memory: float = 20):
        devices = Device.cuda.all()
        devices = {d.cuda_index: d for d in devices}
        self.pool = [CUDAElement(devices[id], min_memory) for id in ids]

    @cached_property
    def allocate(self):
        return BaseAllocator[CUDAElement](self)

    def __repr__(self) -> str:
        return f"CUDAPool(avai={self.available_size})"



import inspect
from typing import Any, Tuple

from .exp import Exp
from .runner.csv import CSVRunner


class ExpFunc:

    def __init__(self, exp_func) -> None:
        self.exp_func = exp_func
        self.run_csv = CSVRunner.set(self).run

    async def __call__(self, exp: Exp, **kwargs) -> Tuple[Exp, Any]:
        assert isinstance(exp, Exp)

        # filter out kwargs that are not needed
        need_kwargs = inspect.signature(self.exp_func).parameters.keys()
        filtered_kwargs = {k: v for k, v in kwargs.items() if k in need_kwargs}
        ba = inspect.signature(self.exp_func).bind(exp, **filtered_kwargs)
        ba.apply_defaults()

        results = await self.exp_func(**ba.arguments)
        await exp.cleanup()
        return (exp, results)


def exp_func(func):
    return ExpFunc(func)



import asyncio
import os
import shutil
from functools import cached_property
from logging import getLogger
from pathlib import Path
from typing import List, Literal, Optional

import psutil

from ..threads import to_thread
from .base import BaseAllocator, BaseElement, BasePool, BaseResources

logger = getLogger(__name__)


class DiskElement(BaseElement):

    def __init__(self, size: int, source_folder: str, target_folder: str,
                 cleanup_target: bool, disk_allocator: "DiskAllocator"):
        self.size = size
        self.source_folder = source_folder
        self.target_folder = target_folder
        self.cleanup_target = cleanup_target
        self.disk_allocator = disk_allocator
        self.allocated = True

    async def cleanup(self):
        self.allocated = False
        if self.cleanup_target:
            logger.info(f"Cleaning up {self.target_folder}")
            await to_thread(shutil.rmtree, self.disk_allocator.target_folder)


class DiskAllocator(BaseAllocator[DiskElement]):

    pool: "DiskPool"
    source_folder: str
    target_folder: str
    cleanup_target: bool = True

    async def _allocate(self, size: int):
        if size > self.pool.available_size:
            return []

        self.pool.pre_allocated += size
        return [
            DiskElement(size, self.source_folder, self.target_folder,
                        self.cleanup_target, self)
        ]


class CopyAllocator(DiskAllocator):

    def __init__(self, pool, unit, max_copys=2):
        super().__init__(pool)
        self.in_copy = 0
        self.max_copys = max_copys
        self.unit = unit

    async def _callback(
        self,
        _allocated: BaseResources,
        source_folder: str,
        target_folder: str,
        files: Optional[List[str]] = None,
        cleanup_target: bool = True,
    ):

        def copy_files(
            source_dir: Path,
            target_dir: Path,
            files: Optional[List[str]] = None,
        ):
            for f in files:
                if not (target_dir / f).exists() or (
                        source_dir / f).stat().st_size != (target_dir /
                                                           f).stat().st_size:
                    src = (source_dir / f).as_posix()
                    tgt = (target_dir / f).as_posix()
                    shutil.copyfile(src, tgt)

        self.source_folder = source_folder
        self.target_folder = target_folder
        self.cleanup_target = cleanup_target
        source_dir = Path(source_folder)
        target_dir = Path(target_folder)
        if files is None:
            files = [
                f for f in os.listdir(source_folder)
                if os.path.isfile(source_dir / f)
            ]

        while self.in_copy >= self.max_copys:
            await asyncio.sleep(1)

        self.in_copy += 1
        await to_thread(copy_files, source_dir, target_dir, files)
        self.pool.pre_allocated -= _allocated.size()
        self.in_copy -= 1

    async def _get_size(
        self,
        source_folder: str,
        target_folder: str,
        files: Optional[List[str]] = None,
        cleanup_target: bool = True,
    ):
        source_dir = Path(source_folder)
        if files is None:
            files = [
                f for f in os.listdir(source_folder)
                if os.path.isfile(source_dir / f)
            ]

        size = sum((source_dir / f).stat().st_size for f in files) // self.unit

        target_dir = Path(target_folder)
        if not target_dir.exists():
            target_dir.mkdir(parents=True, exist_ok=True)
        else:
            size -= sum(
                (target_dir / f).stat().st_size
                for f in files if (target_dir / f).exists()) // self.unit
        return size


class DiskPool(BasePool):

    unit_mapping = {'GB': 1_000_000_000, 'MB': 1_000_000}

    def __init__(
        self,
        path: str,
        unit: Literal['GB', 'MB'] = 'GB',
        max_copys: int = 2,
    ):
        self.path = path
        self.unit = self.unit_mapping[unit]
        self.pre_allocated = 0
        self.callback_count = 0
        self.max_copys = max_copys

    @cached_property
    def allocate(self):
        return DiskAllocator(self)

    @cached_property
    def copy_folder(self):
        """Base allocator.

        Args:
            source_folder: Source folder to copy files from.
            target_folder: Target folder to copy files to.
            files: List of files to copy. If None, all files in source_folder
                will be copied.
        """
        return CopyAllocator(self, max_copys=self.max_copys, unit=self.unit)

    @property
    def available_size(self) -> int:
        return psutil.disk_usage(
            self.path).free // self.unit - self.pre_allocated



import asyncio
from logging import getLogger
from typing import TYPE_CHECKING, Any, Dict

from typing_extensions import Self

if TYPE_CHECKING:
    from ..func import ExpFunc

logger = getLogger(__name__)


class BaseRunner:

    exp_func: "ExpFunc"

    @classmethod
    def set(cls, exp_func: "ExpFunc") -> "Self":

        class RunnerWithExpFunc(cls):

            def __new__(cls) -> Self:
                self = super().__new__(cls)
                self.exp_func = exp_func
                return self

        return RunnerWithExpFunc()

    def create_task(self, uuid: str, **kwargs):
        from ..exp import Exp
        logger.info(f"Create task: {uuid}")
        return asyncio.create_task(self.exp_func(Exp(self, uuid), **kwargs),
                                   name=uuid)

    def run(self, *args, **kwargs):
        raise NotImplementedError("run method is not implemented")

    async def arun(self, *args, **kwargs):
        raise NotImplementedError("run method is not implemented")

    async def _report(self, uuid: str, metrics: Dict[str, Any]):
        raise NotImplementedError("_report method is not implemented")
